"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.explainPeerRequirement = explainPeerRequirement;
exports.explainPeerRequirements = explainPeerRequirements;
const tslib_1 = require("tslib");
const cli_1 = require("@yarnpkg/cli");
const core_1 = require("@yarnpkg/core");
const clipanion_1 = require("clipanion");
const t = tslib_1.__importStar(require("typanion"));
// eslint-disable-next-line arca/no-default-export
class ExplainPeerRequirementsCommand extends cli_1.BaseCommand {
    constructor() {
        super(...arguments);
        this.hash = clipanion_1.Option.String({
            required: false,
            validator: t.cascade(t.isString(), [
                t.matchesRegExp(/^p[0-9a-f]{5}$/),
            ]),
        });
    }
    async execute() {
        const configuration = await core_1.Configuration.find(this.context.cwd, this.context.plugins);
        const { project } = await core_1.Project.find(configuration, this.context.cwd);
        await project.restoreInstallState({
            restoreResolutions: false,
        });
        await project.applyLightResolution();
        if (typeof this.hash !== `undefined`) {
            return await explainPeerRequirement(this.hash, project, {
                stdout: this.context.stdout,
            });
        }
        else {
            return await explainPeerRequirements(project, {
                stdout: this.context.stdout,
            });
        }
    }
}
ExplainPeerRequirementsCommand.paths = [
    [`explain`, `peer-requirements`],
];
ExplainPeerRequirementsCommand.usage = clipanion_1.Command.Usage({
    description: `explain a set of peer requirements`,
    details: `
      A peer requirement represents all peer requests that a subject must satisfy when providing a requested package to requesters.

      When the hash argument is specified, this command prints a detailed explanation of the peer requirement corresponding to the hash and whether it is satisfied or not.

      When used without arguments, this command lists all peer requirements and the corresponding hash that can be used to get detailed information about a given requirement.

      **Note:** A hash is a six-letter p-prefixed code that can be obtained from peer dependency warnings or from the list of all peer requirements (\`yarn explain peer-requirements\`).
    `,
    examples: [[
            `Explain the corresponding peer requirement for a hash`,
            `$0 explain peer-requirements p1a4ed`,
        ], [
            `List all peer requirements`,
            `$0 explain peer-requirements`,
        ]],
});
exports.default = ExplainPeerRequirementsCommand;
async function explainPeerRequirement(peerRequirementsHash, project, opts) {
    const root = project.peerRequirementNodes.get(peerRequirementsHash);
    if (typeof root === `undefined`)
        throw new Error(`No peerDependency requirements found for hash: "${peerRequirementsHash}"`);
    const seen = new Set();
    const makeTreeNode = (request) => {
        if (seen.has(request.requester.locatorHash)) {
            return {
                value: core_1.formatUtils.tuple(core_1.formatUtils.Type.DEPENDENT, { locator: request.requester, descriptor: request.descriptor }),
                children: request.children.size > 0
                    ? [{ value: core_1.formatUtils.tuple(core_1.formatUtils.Type.NO_HINT, `...`) }]
                    : [],
            };
        }
        seen.add(request.requester.locatorHash);
        return {
            value: core_1.formatUtils.tuple(core_1.formatUtils.Type.DEPENDENT, { locator: request.requester, descriptor: request.descriptor }),
            children: Object.fromEntries(Array.from(request.children.values(), child => {
                return [
                    core_1.structUtils.stringifyLocator(child.requester),
                    makeTreeNode(child),
                ];
            })),
        };
    };
    const warning = project.peerWarnings.find(warning => {
        return warning.hash === peerRequirementsHash;
    });
    const report = await core_1.StreamReport.start({
        configuration: project.configuration,
        stdout: opts.stdout,
        includeFooter: false,
        includePrefix: false,
    }, async (report) => {
        const Marks = core_1.formatUtils.mark(project.configuration);
        const mark = warning ? Marks.Cross : Marks.Check;
        report.reportInfo(core_1.MessageName.UNNAMED, `Package ${core_1.formatUtils.pretty(project.configuration, root.subject, core_1.formatUtils.Type.LOCATOR)} is requested to provide ${core_1.formatUtils.pretty(project.configuration, root.ident, core_1.formatUtils.Type.IDENT)} by its descendants`);
        report.reportSeparator();
        report.reportInfo(core_1.MessageName.UNNAMED, core_1.formatUtils.pretty(project.configuration, root.subject, core_1.formatUtils.Type.LOCATOR));
        core_1.treeUtils.emitTree({
            children: Object.fromEntries(Array.from(root.requests.values(), request => {
                return [
                    core_1.structUtils.stringifyLocator(request.requester),
                    makeTreeNode(request),
                ];
            })),
        }, {
            configuration: project.configuration,
            stdout: opts.stdout,
            json: false,
        });
        report.reportSeparator();
        if (root.provided.range === `missing:`) {
            const problem = warning ? `` : ` , but all peer requests are optional`;
            report.reportInfo(core_1.MessageName.UNNAMED, `${mark} Package ${core_1.formatUtils.pretty(project.configuration, root.subject, core_1.formatUtils.Type.LOCATOR)} does not provide ${core_1.formatUtils.pretty(project.configuration, root.ident, core_1.formatUtils.Type.IDENT)}${problem}.`);
        }
        else {
            const providedLocatorHash = project.storedResolutions.get(root.provided.descriptorHash);
            if (!providedLocatorHash)
                throw new Error(`Assertion failed: Expected the descriptor to be registered`);
            const providedPackage = project.storedPackages.get(providedLocatorHash);
            if (!providedPackage)
                throw new Error(`Assertion failed: Expected the package to be registered`);
            report.reportInfo(core_1.MessageName.UNNAMED, `${mark} Package ${core_1.formatUtils.pretty(project.configuration, root.subject, core_1.formatUtils.Type.LOCATOR)} provides ${core_1.formatUtils.pretty(project.configuration, root.ident, core_1.formatUtils.Type.IDENT)} with version ${core_1.structUtils.prettyReference(project.configuration, providedPackage.version ?? `0.0.0`)}, ${warning ? `which does not satisfy all requests.` : `which satisfies all requests`}`);
            if (warning?.type === core_1.PeerWarningType.NodeNotCompatible) {
                if (warning.range) {
                    report.reportInfo(core_1.MessageName.UNNAMED, `  The combined requested range is ${core_1.formatUtils.pretty(project.configuration, warning.range, core_1.formatUtils.Type.RANGE)}`);
                }
                else {
                    report.reportInfo(core_1.MessageName.UNNAMED, `  Unfortunately, the requested ranges have no overlap`);
                }
            }
        }
    });
    return report.exitCode();
}
async function explainPeerRequirements(project, opts) {
    const report = await core_1.StreamReport.start({
        configuration: project.configuration,
        stdout: opts.stdout,
        includeFooter: false,
        includePrefix: false,
    }, async (report) => {
        const Marks = core_1.formatUtils.mark(project.configuration);
        const sorted = core_1.miscUtils.sortMap(project.peerRequirementNodes, [
            ([, requirement]) => core_1.structUtils.stringifyLocator(requirement.subject),
            ([, requirement]) => core_1.structUtils.stringifyIdent(requirement.ident),
        ]);
        for (const [, peerRequirement] of sorted.values()) {
            if (!peerRequirement.root)
                continue;
            const warning = project.peerWarnings.find(warning => {
                return warning.hash === peerRequirement.hash;
            });
            const allRequests = [...core_1.structUtils.allPeerRequests(peerRequirement)];
            let andOthers;
            if (allRequests.length > 2)
                andOthers = ` and ${allRequests.length - 1} other dependencies`;
            else if (allRequests.length === 2)
                andOthers = ` and 1 other dependency`;
            else
                andOthers = ``;
            if (peerRequirement.provided.range !== `missing:`) {
                const providedResolution = project.storedResolutions.get(peerRequirement.provided.descriptorHash);
                if (!providedResolution)
                    throw new Error(`Assertion failed: Expected the resolution to have been registered`);
                const providedPkg = project.storedPackages.get(providedResolution);
                if (!providedPkg)
                    throw new Error(`Assertion failed: Expected the provided package to have been registered`);
                const message = `${core_1.formatUtils.pretty(project.configuration, peerRequirement.hash, core_1.formatUtils.Type.CODE)} → ${warning ? Marks.Cross : Marks.Check} ${core_1.structUtils.prettyLocator(project.configuration, peerRequirement.subject)} provides ${core_1.structUtils.prettyLocator(project.configuration, providedPkg)} to ${core_1.structUtils.prettyLocator(project.configuration, allRequests[0].requester)}${andOthers}`;
                if (warning) {
                    report.reportWarning(core_1.MessageName.UNNAMED, message);
                }
                else {
                    report.reportInfo(core_1.MessageName.UNNAMED, message);
                }
            }
            else {
                const message = `${core_1.formatUtils.pretty(project.configuration, peerRequirement.hash, core_1.formatUtils.Type.CODE)} → ${warning ? Marks.Cross : Marks.Check} ${core_1.structUtils.prettyLocator(project.configuration, peerRequirement.subject)} doesn't provide ${core_1.structUtils.prettyIdent(project.configuration, peerRequirement.ident)} to ${core_1.structUtils.prettyLocator(project.configuration, allRequests[0].requester)}${andOthers}`;
                if (warning) {
                    report.reportWarning(core_1.MessageName.UNNAMED, message);
                }
                else {
                    report.reportInfo(core_1.MessageName.UNNAMED, message);
                }
            }
        }
    });
    return report.exitCode();
}
